;;; ----------------------------------------------------------------------------
;;; $RCSfile: Render.lisp,v $
;;; $Author: hubert $
;;; $Date: 2000/05/05 06:13:39 $
;;; $Revision: 1.1.1.1 $
;;; ----------------------------------------------------------------------------

;;; Drawing routines that affect the framebuffer (*imagebuf*)

(declaim (optimize (speed 3)
		   (compilation-speed 0)
		   (safety 0)
		   (debug 0)))


(defmacro swap (a b)
  (let ((tmp (gensym)))
    `(prog ((,tmp ,a))
	   (setf ,a ,b)
	   (setf ,b ,tmp))))


(declaim (inline setpix ClearColor))


(defun rgb-to-pixel-32 (r g b)
  (declare (type card8 r g b))
  (logior (logior (ash  r 16) (ash g 8)) b))


(defun rgb-to-pixel-16 (r g b)
  (declare (type card8 r g b))
  (logior (logior (ash  (ash r -3) 11) (ash (ash g -2) 5)) (ash b -3)))


(defun ClearColor (r g b)
  (setf *clear-color* (rgb-to-pixel-16 r g b)))


(defmacro make-clear-fun (funname &rest body)
  `(defun ,funname ()
     (do ((y 0 (1+ y)))
	 ((>= y *screen-height*))
       (declare (type card16 y))
       (do ((x 0 (1+ x)))
	   ((>= x *screen-width*))
	 (declare (type card16 x))
	 ,@body))))


(make-clear-fun clear-color-buffer (setf (aref *imagebuf* y x) *clear-color*))
(make-clear-fun clear-depth-buffer (setf (aref *zbuffer* y x) 0))
(make-clear-fun clear-color-and-depth-buffers
		(setf (aref *imagebuf* y x) *clear-color*)
		(setf (aref *zbuffer* y x) 0))


(defmacro Clear (&rest buffers)
  (if (and (member :COLOR-BUFFER buffers) (member :DEPTH-BUFFER buffers))
      `(clear-color-and-depth-buffers)
    (if (member :COLOR-BUFFER buffers)
	`(clear-color-buffer)
      (if (member :DEPTH-BUFFER buffers)
	  `(clear-depth-buffer)
	(error "Clear: unknown buffer type")))))


(defun setpix (x y r g b)
  (declare (type card16 x y)
	   (type card8 r g b))
  (let ((color (rgb-to-pixel-16 r g b)))
    (setf (aref *imagebuf* y x) color)))


(defun setpixz (x y z r g b)
  (declare (type card16 x y)
	   (type int32 z)
	   (type card8 r g b))
  (let ((color (rgb-to-pixel-16 r g b)))
    (when (>= z 0)
      (setf z (the int32 (- (the int32 z) (the int32 *far*))))
      (when (<= z (aref *zbuffer* y x))
	(setf (aref *imagebuf* y x) color
	      (aref *zbuffer* y x) z)))
    nil))


(defun rgbline (x1 y1 x2 y2 r g b)
  "Draw line without clipping using Luca algorithm."
  (declare (type int32 x1 y1 x2 y2)
  	   (type card8 r g b))
  ;; begin push clipping
  (when (< x1 0) (setf x1 0))
  (when (< x2 0) (setf x2 0))
  (when (< y1 0) (setf y1 0))
  (when (< y2 0) (setf y2 0))
  
  (when (>= x1 *screen-width*) (setf x1 (- *screen-width* 1)))
  (when (>= x2 *screen-width*) (setf x2 (- *screen-width* 1)))
  (when (>= y1 *screen-height*) (setf y1 (- *screen-height* 1)))
  (when (>= y2 *screen-height*) (setf y2 (- *screen-height* 1)))
  ;; end push clipping

  (let ((x x1)
	(y y1)
	(dx (- x2 x1))
	(dy (- y2 y1))
	(adx 0)
	(ady 0)
	(incx 0)
	(incy 0)
	(cumul 0))
    (declare (type int32 x y)
	     (type int32 dx dy adx ady incx incy cumul))

    ;; print the last point
    (setpix x2 y2 r g b)
	      
    ;; compute increment
    (if (< dx 0)
	(setf adx (- dx) incx -1)
      (setf adx dx incx 1))
	      
    (if (< dy 0)
	(setf ady (- dy) incy -1)
      (setf ady dy incy 1))
	      
    ;; start draw line
    (if (> adx ady)
	(progn
	  (setf cumul (ash adx -1))
	  (do ((i 0 (1+ i)))
	      ((>= i adx))
	    (declare (type card16 i))
	    (setpix x y r g b)
	    (incf x incx)
	    (incf cumul ady)
	    (when (>= cumul adx)
	      (decf cumul adx)
	      (incf y incy))))
      (progn
	(setf cumul (ash ady -1))
	(do ((i 0 (1+ i)))
	    ((>= i ady))
	  (declare (type card16 i))
	  (setpix x y r g b)
	  (incf y incy)
	  (incf cumul adx)
	  (when (>= cumul ady)
	    (decf cumul ady)
	    (incf x incx)))))))



;;polygon-filling functions
(macrolet
    ;;function-specific macros   
    ((fillpoly3-do-action1 () '(line_h (ash xd -8) (ash xf -8) l))
     (fillpoly3-do-action2 () '(line_h (ash xf -8) (ash xd -8) l))
	   
     (fillpoly3gour-do-action1 () '(progn (line_hg (ash xd -8) (ash rd -8) (ash gd -8) (ash bd -8)
						   (ash xf -8) (ash rf -8) (ash gf -8) (ash bf -8) l)
					  (list-match-incf (rd  gd  bd  rf  gf  bf)
							   (d1r d1g d1b d2r d2g d2b))))
     
     (fillpoly3gour-do-action2 () '(progn (line_hg (ash xf -8) (ash rf -8) (ash gf -8) (ash bf -8)
						   (ash xd -8) (ash rd -8) (ash gd -8) (ash bd -8) l)
					  (list-match-incf (rd  gd  bd  rf  gf  bf)
							   (d1r d1g d1b d2r d2g d2b))))
     
     (fillpoly3z-do-action1 () '(progn (line_hz (ash xd -8) (ash xf -8)
						(ash zd -8) (ash zf -8) l)
				       (incf zd deltaz1)
				       (incf zf deltaz2)))
     
     (fillpoly3z-do-action2 () '(progn (line_hz (ash xf -8) (ash xd -8)
						(ash zf -8) (ash zd -8) l)
				       (incf zd deltaz1)
				       (incf zf deltaz2)))
     
     (fillpoly3zgour-do-action1 () '(progn (line_hzg (ash xd -8) (ash zd -8) (ash rd -8) (ash gd -8) (ash bd -8)
						     (ash xf -8) (ash zf -8) (ash rf -8) (ash gf -8) (ash bf -8) l)
					   (list-match-incf (zd      zf      rd  gd  bd  rf  gf  bf)
							    (deltaz1 deltaz2 d1r d1g d1b d2r d2g d2b))))

     (fillpoly3zgour-do-action2 () '(progn (line_hzg (ash xf -8) (ash zf -8) (ash rf -8) (ash gf -8) (ash bf -8)
						     (ash xd -8) (ash zd -8) (ash rd -8) (ash gd -8) (ash bd -8) l)
					   (list-match-incf (zd      zf      rd  gd  bd  rf  gf  bf)
							    (deltaz1 deltaz2 d1r d1g d1b d2r d2g d2b)))))
  
  (macrolet
      ;;common macros      
      ((abc-swap (&body swapground)
		 (let* ((abc-comb '((a b) (a c) (b c)))
			(swapcode (lambda (x a b) `(swap ,(symb x a) ,(symb x b))))
			(whencode (lambda (a b) `(when (> ,(symb 'Y a) ,(symb 'Y b))
						   ,@(mapcar (lcurry swapcode nil a b) swapground))))
			(mapwhen (lambda (l) (funcall whencode (car l) (cadr l)))))
		   `(progn ,@(mapcar mapwhen abc-comb))))
	     
	     
       (if-delta (dnum1 dnum2 ynum1 ynum2 do-action1 do-action2)
		 (let* ((delta1 (symb 'DELTA dnum1))
			(delta2 (symb 'DELTA dnum2))
			(y1 (symb 'Y ynum1))
			(y2 (symb 'Y ynum2))
			(gen-do (lambda (do-action)
				  `(do ((l ,y1 (1+ l)))
				       ((>= l ,y2))
				     (declare (type int32 l))
				     (,do-action)
				     (incf xd delta1)
				     (incf xf delta2)))))
		   `(if (< ,delta1 ,delta2)
			,(funcall gen-do do-action1)
		      ,(funcall gen-do do-action2)))))
    
    (flet
	;;common functions	
	((when-delta1 (x1 x2 y1 y2)
		      (truncate (the int32 (ash (the int32 (- x1 x2)) 8))
				(the int32 (- y1 y2))))
	 
	 (when-d (r1 r2 y1 y2) (truncate (ash (- r1 r2) 8) (the int32 (- y1 y2)))))
      (declare (inline when-delta1)
	       (inline when-d))

      
      ;;FIRST DEFINED FUNCTION 
      (defun fillpoly3 (xa ya xb yb xc yc r g b)
	"Fill a triangle with uniform color."
	(declare (type int32 xa ya xb yb xc yc)
		 (type card8 r g b))
	
	(flet ((line_h (xd xf h)
		       (declare (type int32 xd xf h))
		       ;; clipping haut et bas
		       (when (or (< h 0) (>= h *screen-height*)) (return-from line_h))
		       ;; clipping gauche
		       (when (< xd 0) (setf xd 0))
		       ;; clipping droit
		       (when (>= xf *screen-width*) (setf xf (- *screen-width* 1)))
		       
		       (do ((i xd (1+ i)))
			   ((> i xf))
			 (declare (type card16 i))
			 (setpix i h r g b))))
	  
	  
	  (let ((delta1 0)
		(delta2 0)
		(xd 0)
		(xf 0))
	    (declare (type int32 delta1 delta2 xd xf))
	    
	    ;; trie des y
	    (abc-swap X Y)
	    
	    ;; remplissage
	    (when (/= ya yc) (setf delta1 (the int32 (when-delta1 xa xc ya yc))))
	    (when (/= ya yb) (setf delta2 (the int32 (when-delta1 xb xa yb ya))))
	    
	    (setf xd (ash (the int32  xa) 8))
	    (setf xf xd)
	    
	    (if-delta 1 2 A B fillpoly3-do-action1 fillpoly3-do-action2)
	    
	    (when (/= yb yc) (setf delta2 (the int32 (when-delta1 xb xc yb yc)))
		  (setf xf (ash (the int32 xb) 8))
		  (if-delta 2 1 B C fillpoly3-do-action1 fillpoly3-do-action2)))))
	  
      
      ;;SECOND DEFINED FUNCTION
      (defun fillpoly3gour (xa ya ra ga ba xb yb rb gb bb xc yc rc gc bc)
	"Fill a triangle with gouraud shading."
	(declare (type int32 xa ya xb yb xc yc)
		 (type card8 ra ga ba rb gb bb rc gc bc))
	
	(flet ((line_hg (xd rd gd bd xf rf gf bf h)
			(declare (type int32 xd xf h)
				 (type int32 rd gd bd rf gf bf))
			(let ((ir 0)
			      (ig 0)
			      (ib 0)
			      (cr (ash rd 8))
			      (cg (ash gd 8))
			      (cb (ash bd 8)))
			  (declare (type int32 ir ig ib cr cg cb))
			  ;; clipping haut et bas
			  (when (or (< h 0) (>= h *screen-height*)) (return-from line_hg))
			  
			  (if (= xd xf)
			      (setf ir 0
				    ig 0
				    ib 0)
			    (setf ir (when-delta1 rf rd xf xd)
				  ig (when-delta1 gf gd xf xd)
				  ib (when-delta1 bf bd xf xd)))
			    
			  ;; clipping droit
			  (when (>= xf *screen-width*) (setf xf (- *screen-width* 1)))
			  (do ((i xd (1+ i)))
			      ((> i xf))
			    (declare (type int32 i))
			    (if (>= i 0) (setpix i h
						 (the card8 (ash cr -8))
						 (the card8 (ash cg -8))
						 (the card8 (ash cb -8))))
			    (incf cr ir)
			    (incf cg ig)
			    (incf cb ib)))))
		 
	  
	  (let ((delta1 0)
		(delta2 0)
		(d1r 0) (d1g 0) (d1b 0) ; delta1 color
		(d2r 0) (d2g 0) (d2b 0) ; delta2 color
		(rd 0) (gd 0) (bd 0)	; debut color
		(rf 0) (gf 0) (bf 0)	; fin color
		(xd 0)
		(xf 0))
	    
	    (declare (type int32 delta1 delta2
			   d1r d1g d1b
			   d2r d2g d2b
			   rd gd bd
			   rf gf bf
			   xd xf))
	    
	    ;; trie des y
	    (abc-swap X Y R G B)
	    
	    ;; remplissage
	    (when (/= ya yc)
	      (setf delta1 (the int32 (when-delta1 xa xc ya yc))
		    d1r (when-d ra rc ya yc)
		    d1g (when-d ga gc ya yc)
		    d1b (when-d ba bc ya yc)))
	    
	    (when (/= ya yb)
	      (setf delta2 (the int32 (when-delta1 xb xa yb ya))
		    d2r (when-d rb ra yb ya)
		    d2g (when-d gb ga yb ya) 
		    d2b (when-d bb ba yb ya)))
	    
	    (setf xd (ash (the int32  xa) 8))
	    (setf rd (ash ra 8))
	    (setf gd (ash ga 8))
	    (setf bd (ash ba 8))
	    (list-match-setf (xf rf gf bf) (xd rd gd bd))
	    	    
	    (if-delta 1 2 A B fillpoly3gour-do-action1 fillpoly3gour-do-action2)
	    
	    (when (/= yb yc)
	      (setf delta2 (the int32 (when-delta1 xb xc yb yc)))
	      (setf xf (ash (the int32 xb) 8))
	      (setf d2r (when-d rb rc yb yc))
	      (setf d2g (when-d gb gc yb yc))
	      (setf d2b (when-d bb bc yb yc))
	      (setf rf (ash rb 8))
	      (setf gf (ash gb 8))
	      (setf bf (ash bb 8))

	      (if-delta 2 1 B C fillpoly3gour-do-action1 fillpoly3gour-do-action2)))))


      ;;THIRD DEFINED FUNCTION
      (defun fillpoly3z (xa ya za xb yb zb xc yc zc r g b)
	"fill a triangle with uniform color and zbuffer"
	(declare (type int32 xa ya za xb yb zb xc yc zc)
		 (type card8 r g b))
	(flet ((line_hz (xd xf zd zf h)
			(declare (type int32 xd zd xf zf h))
			(let ((deltaz 0);; delta z
			      (curz 0));; current z
			  (declare (type int32 deltaz curz))
			  ;; clipping haut et bas
			  (when (or (< h 0) (>= h *screen-height*)) (return-from line_hz))
			  ;; clipping gauche
			  (when (< xd 0) (setf xd 0))
			  ;; clipping droit
			  (when (>= xf *screen-width*) (setf xf (- *screen-width* 1)))
			  ;; erreur de sens
			  (when (> xd xf) (return-from line_hz))

			  (when (/= xd xf)
			    (setf deltaz (when-delta1 zf zd xf xd)))
			  (setf curz (ash zd 8))
			  
			  (do ((i xd (1+ i)))
			      ((> i xf))
			    (declare (type card16 i))
			    (setpixz i h (ash curz -8) r g b)
			    (incf curz deltaz)))))
	  
	  (let ((delta1 0)
		(delta2 0)
		(deltaz1 0)
		(deltaz2 0)
		(xd 0)
		(xf 0)
		(zd 0)
		(zf 0))
	    (declare (type int32 delta1 delta2 deltaz1 deltaz2 xd xf zd zf))

	    ;; trie des y
	    (abc-swap x y z)
	    
	    ;; remplissage
	    (when (/= ya yc)
	      (setf delta1 (the int32 (when-delta1 xa xc ya yc))
		    deltaz1 (the int32 (when-delta1 za zc ya yc))))

	    (when (/= ya yb)
	      (setf delta2 (the int32 (when-delta1 xb xa yb ya))
		    deltaz2 (the int32 (when-delta1 zb za yb ya))))

	    (setf xd (ash (the int32 xa) 8))
	    (setf xf xd)
	    (setf zd (ash (the int32 za) 8))
	    (setf zf zd)

	    (if-delta 1 2 A B fillpoly3z-do-action1 fillpoly3z-do-action2)
	    
	    (when (/= yb yc)
	      (setf delta2 (the int32 (when-delta1 xb xc yb yc)))
	      (setf deltaz2 (the int32 (when-delta1 zb zc yb yc)))
	      (setf xf (ash (the int32 xb) 8))
	      (setf zf (ash (the int32 zb) 8))

	      (if-delta 2 1 B C fillpoly3z-do-action1 fillpoly3z-do-action2)))))
	      
      
      ;;FOURTH DEFINED FUNCTION
      (defun fillpoly3zgour (xa ya za ra ga ba
				xb yb zb rb gb bb
				xc yc zc rc gc bc)
	"Fill a triangle with gouraud shading."
	(declare (type int32 xa ya za xb yb zb xc yc zc)
		 (type card8 ra ga ba rb gb bb rc gc bc))
	
	(flet ((line_hzg (xd zd rd gd bd xf zf rf gf bf h)
			 (declare (type int32 xd zd xf zf h)
				  (type int32 rd gd bd rf gf bf))
			 (let ((deltaz 0)
			       (curz 0)
			       (ir 0)
			       (ig 0)
			       (ib 0)
			       (cr (ash rd 8))
			       (cg (ash gd 8))
			       (cb (ash bd 8)))
			   (declare (type int32 deltaz curz ir ig ib cr cg cb))
			   ;; clipping haut et bas
			   (when (or (< h 0) (>= h *screen-height*)) (return-from line_hzg))
			   ;; erreur de sens
			   (when (> xd xf) (return-from line_hzg))
		     
			   (when (/= xd xf)
			     (setf deltaz (when-delta1 zf zd xf xd)
				   ir (when-delta1 rf rd xf xd)
				   ig (when-delta1 gf gd xf xd)
				   ib (when-delta1 bf bd xf xd)))

			   (setf curz (ash zd 8))
			   ;; clipping droit
			   (when (>= xf *screen-width*) (setf xf (- *screen-width* 1)))
			   (do ((i xd (1+ i)))
			       ((> i xf))
			     (declare (type int32 i))
			     (if (>= i 0) (setpixz i h (ash curz -8)
						   (the card8 (ash cr -8))
						   (the card8 (ash cg -8))
						   (the card8 (ash cb -8))))
			     (incf curz deltaz)
			     (incf cr ir)
			     (incf cg ig)
			     (incf cb ib)))))
    
	  (let ((delta1 0)
		(delta2 0)
		(deltaz1 0)
		(deltaz2 0)
		(d1r 0) (d1g 0) (d1b 0) ; delta1 color
		(d2r 0) (d2g 0) (d2b 0) ; delta2 color
		(rd 0) (gd 0) (bd 0)	; debut color
		(rf 0) (gf 0) (bf 0)	; fin color
		(xd 0)
		(xf 0)
		(zd 0)
		(zf 0))
	    
	    (declare (type int32 delta1 delta2
			   deltaz1 deltaz2
			   d1r d1g d1b
			   d2r d2g d2b
			   rd gd bd
			   rf gf bf
			   xd xf zd zf))
	    
	    ;; trie des y
	    (abc-swap X Y R G B Z)
	    
	    ;; remplissage
	    (when (/= ya yc)
	      (setf delta1 (the int32 (when-delta1 xa xc ya yc))
		    deltaz1 (the int32 (when-delta1 za zc ya yc))
		    d1r (when-d ra rc ya yc)
		    d1g (when-d ga gc ya yc) 
		    d1b (when-d ba bc ya yc)))
	    
	    (when (/= ya yb)
	      (setf delta2 (the int32 (when-delta1 xb xa yb ya))
		    deltaz2 (the int32 (when-delta1 zb za yb ya))
		    d2r (when-d rb ra yb ya)
		    d2g (when-d gb ga yb ya) 
		    d2b (when-d bb ba yb ya)))
    
	    (setf xd (ash (the int32 xa) 8))
	    (setf zd (ash (the int32 za) 8))
	    (setf rd (ash ra 8))
	    (setf gd (ash ga 8))
	    (setf bd (ash ba 8))
	    (list-match-setf (xf zf rf gf bf) (xd zd rd gd bd))
	    
	    (if-delta 1 2 A B fillpoly3zgour-do-action1 fillpoly3zgour-do-action2)	    

	    (when (/= yb yc)
	      (setf delta2 (the int32 (when-delta1 xb xc yb yc)))
	      (setf deltaz2 (the int32 (when-delta1 zb zc yb yc)))
	      (setf xf (ash (the int32 xb) 8))
	      (setf zf (ash (the int32 zb) 8))
	      (setf d2r (when-d rb rc yb yc))
	      (setf d2g (when-d gb gc yb yc))
	      (setf d2b (when-d bb bc yb yc))
	      (setf rf (ash rb 8))
	      (setf gf (ash gb 8))
	      (setf bf (ash bb 8))

	      (if-delta 2 1 B C fillpoly3zgour-do-action1 fillpoly3zgour-do-action2))))))))
