;;; ----------------------------------------------------------------------------
;;; $RCSfile: VertexBuffer.lisp,v $
;;; $Author: hubert $
;;; $Date: 2000/05/05 06:13:40 $
;;; $Revision: 1.1.1.1 $
;;; ----------------------------------------------------------------------------

;;GLOS (Graphic Library in Open Source), an ANSI Common Lisp OpenGL subset.
;;Copyright (C) 2000 the GLOS development team (http://glos.sourceforge.net)

(declaim (optimize (speed 3)
		   (compilation-speed 0)
		   (safety 0)
		   (debug 0)))


(deftype rgb    () '(simple-array card8 (3)))
(deftype pcoord () '(simple-array int16 (3))) ;to hold window coordinates


(defun rgb-make (r g b)
  (declare (type card8 r g b))
  (make-array 3 :element-type 'card8 :initial-contents (list r g b)))


;setable accesors
(defmacro rgb-r (col)
  `(aref ,col 0))
(defmacro rgb-g (col)
  `(aref ,col 1))
(defmacro rgb-b (col)
  `(aref ,col 2))


(declaim (inline rgb-set))
(defun rgb-set (col r g b)
  (declare (type rgb col))
  (declare (type card8 r g b))
  (setf (rgb-r col) r
	(rgb-g col) g
	(rgb-b col) b))


(defun pcoord-make-null ()
  (make-array 3 :element-type 'int16))


(defstruct vertex
  (coords   (vec4-make-null)   :type vec4)      ;object coordinates
  (normal   (vec3-make-null)   :type vec3)      ;normal of vertex
  (material (make-material)    :type material)  ;material of vertex
  (color    (rgb-make 0 0 0)   :type rgb)       ;color of vertex
  (tcoords  (vec4-make-null)   :type vec4)      ;world (eye) coordinates
  (tnormal  (vec3-make-null)   :type vec3)      ;normal after modelview transform
  (pcoords  (pcoord-make-null) :type pcoord))   ;window coordinates


(deftype vb-array-index () 'card16)
(defparameter *rgb-white* (rgb-make 255 255 255))
(defparameter *vb-current-color* *rgb-white*)
(defparameter *vb-current-normal* (vec3-make-null))
;;; vbsize must be divisible by 6 to be able to specify complete lines (divisible by 2)
;;; and complete triangles (divisible by 3)
(defparameter *vb-size* 120)
(defparameter *vb* (make-array *vb-size* :element-type 'vertex))
(defparameter *vb-index* 0)
(defparameter *vb-draw-mode* :UNKNOWN)


(declaim (type (simple-vector *) *vb*))
(declaim (type rgb  *vb-current-color*))
(declaim (type vec3  *vb-current-normal*))
(declaim (type vb-array-index *vb-index* *vb-size*))


;; allocates memory for the vb to avoid consing
(dotimes (i *vb-size*)
  (setf (aref *vb* i) (make-vertex)))


(defmacro set-vb-vertex (i slot-type &rest slot-values)
  "Sets ith vertex slot 'slot-type' to 'slot-values' into the vertex buffer"
  (multiple-value-bind (val1 val2 val3 val4) (list-values slot-values)
    (let* ((4access-list (list 'vec-x 'vec-y 'vec-z 'vec-w))
	   (4vals-list (list val1 val2 val3 val4))
	   (3access-list (butlast 4access-list))
	   (3vals-list (butlast 4vals-list))
	   (vertex-name (lambda (x) (symb 'VERTEX- (symbol-name x))))
	   (setf-gen (lambda (x y vertex)
		       `(setf (,x (,(funcall vertex-name vertex) (aref *vb* ,i))) ,y)))
	   (build-3-setf (lambda (vertex) (mapcar (lcurry setf-gen nil nil vertex)
						  3access-list 3vals-list))) 
	   (build-4-setf (lambda (vertex) (mapcar (lcurry setf-gen nil nil vertex)
						  4access-list 4vals-list))))
      (ccase slot-type
	     (:material
	      `(progn
		 (vec4-cpy (material-ambiant (vertex-material (aref *vb* ,i)))
			   (material-ambiant *current-material*))
		 (vec4-cpy (material-diffuse (vertex-material (aref *vb* ,i)))
			   (material-diffuse *current-material*))
		 (vec4-cpy (material-specular (vertex-material (aref *vb* ,i)))
			   (material-specular *current-material*))
		 (setf (material-shininess (vertex-material (aref *vb* ,i)))
		       (material-shininess *current-material*))
		 (vec4-cpy (material-emission (vertex-material (aref *vb* ,i)))
			   (material-emission *current-material*))))
	     ((:coords :tcoords)
	      `(progn ,@(funcall build-4-setf slot-type)))
	     ((:pcoords :normal :tnormal :color)
	      `(progn ,@(funcall build-3-setf slot-type)))))))
		
     
; x,y,z,w in the body are replaced by #{v 0},#{v 1},...
; (with-vec4 v (+ x y z w)) -> (+ #{v 0} #{v 1} #{v 2} #{v 3} )
(defmacro with-vec4 (v &rest body)
  (let* ((x `#{,v 0})
	 (y `#{,v 1})
	 (z `#{,v 2})
	 (w `#{,v 3})
	 (newbody (substn (list x 'x y 'y z 'z w 'w) body)))
    `(prog ()
	   ,@newbody)))


; same as with-vec4 with only x,y,z
(defmacro with-vec3 (v &rest body)
  (let* ((x `#{,v 0})
	 (y `#{,v 1})
	 (z `#{,v 2})
	 (newbody (substn (list x 'x y 'y z 'z) body)))
    `(prog ()
	   ,@newbody)))

    
(defun vec4-transform (i)
  "Object to world coordinates transformation"
  (with-vec4 (vertex-coords (aref *vb* i))
	(macrolet ((row-by-col (row)
 			    (declare (type (card8) row))
 			    `(+ (the single-float (* (the single-float #[*modelview-matrix* ,row 0]) x))
				(the single-float (* (the single-float #[*modelview-matrix* ,row 1]) y))
 				(the single-float (* (the single-float #[*modelview-matrix* ,row 2]) z))
 				(the single-float (* (the single-float #[*modelview-matrix* ,row 3]) w)))))
        (set-vb-vertex i :tcoords
 		      (row-by-col 0)
 		      (row-by-col 1)
 		      (row-by-col 2)
 		      (row-by-col 3)))))
   
		      
(defun normal-transform (i)
  "Transform normals by the upper-left 3x3 matrix of the modelview matrix for lighting"
  (with-vec3 (vertex-normal (aref *vb* i))
	(macrolet ((row-by-col (row)
			       (declare (type (card8) row))
			       `(+ (the single-float (* (the single-float #[*modelview-matrix* ,row 0]) x))
				   (the single-float (* (the single-float #[*modelview-matrix* ,row 1]) y))
				   (the single-float (* (the single-float #[*modelview-matrix* ,row 2]) z)))))
	  (set-vb-vertex i :tnormal
			 (row-by-col 0)
			 (row-by-col 1)
			 (row-by-col 2)))))
			 

(defun vec4-project (i)
  "World to window coordinates transformation"
  (with-vec4 (vertex-tcoords (aref *vb* i))
	     (macrolet ((row-by-col (row)
				    (declare (type card8 row))
				    `(+ (the single-float (* (the single-float #[*projection-matrix* ,row 0]) x))
					(the single-float (* (the single-float #[*projection-matrix* ,row 1]) y))
					(the single-float (* (the single-float #[*projection-matrix* ,row 2]) z))
					(the single-float (* (the single-float #[*projection-matrix* ,row 3]) w)))))
	       ;xc,yc,zc,wc: clip coordinates of the transformed vertex
	       (let ((xc (row-by-col 0)) 
		     (yc (row-by-col 1))
		     (zc (row-by-col 2))
		     (wc (row-by-col 3)))
		 (declare (type single-float xc yc zc wc))
		 ;perspective division (by wc=-z) and viewport transformation for final window coordinates
		 (set-vb-vertex i :pcoords
		    (+ (the card16 (truncate (the single-float (* *viewport-half-width*  (/ xc wc))))) *viewport-centerx*)
		    (+ (the card16 (truncate (the single-float (* *viewport-half-height* (/ yc wc))))) *viewport-centery*)
		    (the card16 (truncate (the single-float zc))))))))	     


(defun do-transforms ()
  (do ((i 0 (1+ i)))
      ((>= i *vb-index*))
    (declare (type vb-array-index i))
    (vec4-transform i) 
    (normal-transform i) 
    (vec4-project i)))


(let ((sum (vec3-make-null)))
  (declare (type vec3 sum))
  (defun do-lighting ()
    (do ((i 0 (1+ i))
	 (vbi nil))
	((>= i *vb-index*))
      (declare (type vb-array-index i))
      (setf vbi (aref *vb* i))
      (do ((j 0 (1+ j))
	   (r 0.0)
	   (g 0.0)
	   (b 0.0)
	   (vnormal    (vertex-tnormal (aref *vb* i)))
	   (mambiant   (material-ambiant   (vertex-material vbi)))
	   (mdiffuse   (material-diffuse   (vertex-material vbi)))
	   (mspecular  (material-specular  (vertex-material vbi)))
	   (mshininess (material-shininess (vertex-material vbi)))
	   (memission  (material-emission  (vertex-material vbi))))
	  ((>= j *nblight*)
	   (set-vb-vertex i :color
			  (min (the int16 (truncate (the (single-float #.(float most-negative-fixnum)
								       #.(float most-positive-fixnum)) (* r 255)))) 255)
			  (min (the int16 (truncate (the (single-float #.(float most-negative-fixnum)
								       #.(float most-positive-fixnum)) (* g 255)))) 255)
			  (min (the int16 (truncate (the (single-float #.(float most-negative-fixnum)
								       #.(float most-positive-fixnum)) (* b 255)))) 255)))
	
	(declare (type card8 j)
		 (type (single-float #.(float most-negative-fixnum)
				     #.(float most-positive-fixnum)) r g b)
		 (type vec3 vnormal)
		 (type vec4 mambiant mdiffuse mspecular memission)
		 (type single-float mshininess))
	(if (aref *tlightsbool* j)
	    (let* ((lposition  (light-position (aref *tlights* j)))
		   (lambiant   (light-ambiant  (aref *tlights* j)))
		   (ldiffuse   (light-diffuse  (aref *tlights* j)))
		   (lspecular  (light-specular (aref *tlights* j)))
		   (cos-angle1 0.0)
		   (cos-angle2 0.0))
	      (declare (type vec4 lposition lambiant ldiffuse lspecular)
		       (type single-float cos-angle1 cos-angle2))
	      ;; ambiant section
	      (incf r (+ (* #{ lambiant 0 } #{ mambiant 0 }) #{ memission 0 }))
	      (incf g (+ (* #{ lambiant 1 } #{ mambiant 1 }) #{ memission 1 }))
	      (incf b (+ (* #{ lambiant 2 } #{ mambiant 2 }) #{ memission 2 }))
	      ;; diffuse section
	      (setf cos-angle1 (vec-dot-product vnormal lposition))
	      (if (> cos-angle1 0.0)
		  (progn
		    (incf r (* (* #{ mdiffuse 0 } #{ ldiffuse 0 }) cos-angle1))
		    (incf g (* (* #{ mdiffuse 1 } #{ ldiffuse 1 }) cos-angle1))
		    (incf b (* (* #{ mdiffuse 2 } #{ ldiffuse 2 }) cos-angle1))))
	      ;; specular section
	      (vec3-set sum #{lposition 0} #{lposition 1} (+ #{lposition 2} 1.0))
	      (vec-set-normalize sum)
	      (setf cos-angle1 (vec-dot-product vnormal sum))
	      (when (> cos-angle1 0.0)
		(setf cos-angle2 (the single-float (expt (the (single-float 0.0) cos-angle1)
							 (the single-float  mshininess))))
		(incf r (* (* #{ mspecular 0 } #{ lspecular 0 }) cos-angle2))
		(incf g (* (* #{ mspecular 1 } #{ lspecular 1 }) cos-angle2))
		(incf b (* (* #{ mspecular 2 } #{ lspecular 2 }) cos-angle2)))))))))

  
(defun do-normals-normalize ()
  (do ((i 0 (1+ i)))
      ((>= i *vb-index*))
    (vec-set-normalize (vertex-tnormal (aref *vb* i)))))


(defun draw-points ()
  (do ((i 0 (setf i (+ i 1))))
      ((>= i *vb-index*))
    (declare (type vb-array-index i ))
    (let ((x (aref (vertex-pcoords (aref *vb* i)) 0))
	  (y (aref (vertex-pcoords (aref *vb* i)) 1))
	  (r  (aref (vertex-color (aref *vb* i)) 0))
	  (g  (aref (vertex-color (aref *vb* i)) 1))
	  (b  (aref (vertex-color (aref *vb* i)) 2)))
      (setpix x y r g b))))


(defun draw-lines ()
  (do ((i 0 (setf i (+ i 2))))
      ((>= i *vb-index*))
    (declare (type vb-array-index i ))
    (let ((x1 (aref (vertex-pcoords (aref *vb* i)) 0))
	  (y1 (aref (vertex-pcoords (aref *vb* i)) 1))
	  (x2 (aref (vertex-pcoords (aref *vb* (+ i 1))) 0))
	  (y2 (aref (vertex-pcoords (aref *vb* (+ i 1))) 1))
	  (r  (aref (vertex-color (aref *vb* i)) 0))
	  (g  (aref (vertex-color (aref *vb* i)) 1))
	  (b  (aref (vertex-color (aref *vb* i)) 2)))
      (rgbline x1 y1 x2 y2 r g b))))


(defun draw-line-loop ()
  (draw-line-strip)
  ;; last line = last to first point
  (let ((lastline (1- *vb-index*)))
    (rgbline
     (aref (vertex-pcoords (aref *vb* lastline)) 0)
     (aref (vertex-pcoords (aref *vb* lastline)) 1)
     (aref (vertex-pcoords (aref *vb* 0)) 0)
     (aref (vertex-pcoords (aref *vb* 0)) 1)
     (aref (vertex-color (aref *vb* lastline)) 0)
     (aref (vertex-color (aref *vb* lastline)) 1)
     (aref (vertex-color (aref *vb* lastline)) 2))))
     
     
(defun draw-line-strip ()
  (do ((i 0 (incf i)))
      ((>= i (1- *vb-index*)))
    (declare (type vb-array-index i ))
    (let ((x1 (aref (vertex-pcoords (aref *vb* i)) 0))
	  (y1 (aref (vertex-pcoords (aref *vb* i)) 1))
	  (x2 (aref (vertex-pcoords (aref *vb* (+ i 1))) 0))
	  (y2 (aref (vertex-pcoords (aref *vb* (+ i 1))) 1))
	  (r  (aref (vertex-color (aref *vb* i)) 0))
	  (g  (aref (vertex-color (aref *vb* i)) 1))
	  (b  (aref (vertex-color (aref *vb* i)) 2)))
      (rgbline x1 y1 x2 y2 r g b))))


(declaim (inline visible-triangle?))
(defun visible-triangle? (x1 y1 x2 y2 x3 y3)
  (declare (type int16 x1 y1 x2 y2 x3 y3))
  (< (- (* (the int16 (- y2 y1))
	   (the int16 (- x3 x1)))
	(* (the int16 (- y3 y1))
	   (the int16 (- x2 x1)))) 0))

	
(defmacro draw-triangle (triangle-type) 
  (let ((vertex1 
	 (case triangle-type
	   ((simple strip) 'i)
	   (fan    '0)))
	(vertex2 
	 (case triangle-type
	   ((simple strip) '(+ i 1))
	   (fan    'i)))
	(vertex3 
	 (case triangle-type
	   ((simple strip) '(+ i 2))
	   (fan    '(+ i 1))))
	(test
	 (case triangle-type
	   (simple '(((i 0 (incf i 3)))
		     ((>= i *vb-index*))))
	   (strip  '(((i 0 (incf i))
		      (inversevertexe 'nil (setf inversevertexe (not inversevertexe))))
		     ((>= i (- *vb-index* 2)))))
	   (fan    '(((i 1 (incf i)))
		     ((>= i (- *vb-index* 1)))))))
	(body
	 (case triangle-type
	   ((simple fan) '((if (visible-triangle? x1 y1 x2 y2 x3 y3)
			      (if (eq *shade-model* :FLAT)
				  (if *use-zbuffer*
				      (fillpoly3z x1 y1 z1
						  x2 y2 z2
						  x3 y3 z3
						  r1 g1 b1)
				    (fillpoly3 x1 y1
					       x2 y2
					       x3 y3
					       r1 g1 b1))
				(if *use-zbuffer*
				    (fillpoly3zgour x1 y1 z1 r1 g1 b1
						    x2 y2 z2 r2 g2 b2
						    x3 y3 z3 r3 g3 b3)
				  (fillpoly3gour x1 y1 r1 g1 b1
						 x2 y2 r2 g2 b2
						 x3 y3 r3 g3 b3))))))
	   (strip        '((if (and  (not inversevertexe) (visible-triangle? x1 y1 x2 y2 x3 y3))
			      (if (eq *shade-model* :FLAT)
				  (if *use-zbuffer*
				      (fillpoly3zgour x1 y1 z1 r1 g1 b1
						      x2 y2 z2 r2 g2 b2
						      x3 y3 z3 r3 g3 b3)
				    (fillpoly3gour x1 y1 r1 g1 b1
						   x2 y2 r2 g2 b2
						   x3 y3 r3 g3 b3))))
			    
			   (if (and inversevertexe (visible-triangle? x2 y2 x1 y1 x3 y3))
			       (if (eq *shade-model* :FLAT)
				   (if *use-zbuffer*
				       (fillpoly3zgour x2 y2 z2 r2 g2 b2
						       x1 y1 z1 r1 g1 b1
						       x3 y3 z3 r3 g3 b3)
				     (fillpoly3gour x2 y2 r2 g2 b2
						    x1 y1 r1 g1 b1
						    x3 y3 r3 g3 b3)))))))))
    `(do ,@test
	 (declare (type vb-array-index i ))
       (let ((x1 (aref (vertex-pcoords (aref *vb* ,vertex1)) 0))
	     (y1 (aref (vertex-pcoords (aref *vb* ,vertex1)) 1))
	     (z1 (aref (vertex-pcoords (aref *vb* ,vertex1)) 2))
	     
	     (x2 (aref (vertex-pcoords (aref *vb* ,vertex2)) 0))
	     (y2 (aref (vertex-pcoords (aref *vb* ,vertex2)) 1))
	     (z2 (aref (vertex-pcoords (aref *vb* ,vertex2)) 2))

	     (x3 (aref (vertex-pcoords (aref *vb* ,vertex3)) 0))
	     (y3 (aref (vertex-pcoords (aref *vb* ,vertex3)) 1))
	     (z3 (aref (vertex-pcoords (aref *vb* ,vertex3)) 2))
	     
	     (r1  (aref (vertex-color (aref *vb* ,vertex1)) 0))
	     (g1  (aref (vertex-color (aref *vb* ,vertex1)) 1))
	     (b1  (aref (vertex-color (aref *vb* ,vertex1)) 2))

	     (r2  (aref (vertex-color (aref *vb* ,vertex2)) 0))
	     (g2  (aref (vertex-color (aref *vb* ,vertex2)) 1))
	     (b2  (aref (vertex-color (aref *vb* ,vertex2)) 2))

	     (r3  (aref (vertex-color (aref *vb* ,vertex3)) 0))
	     (g3  (aref (vertex-color (aref *vb* ,vertex3)) 1))
	     (b3  (aref (vertex-color (aref *vb* ,vertex3)) 2)))
	 
	 ,@body))))
	

(defun draw-triangles ()
  (draw-triangle simple))


(defun draw-triangle-strip ()
  (draw-triangle strip))


(defun draw-triangle-fan ()
  (draw-triangle fan))


(declaim (inline clear-vb))
(defun clear-vb ()
  "Clear the vb to some default values and put the index at the beginning"
  (setf *vb-index* 0))


;(declaim (inline Color))
(defun Color (r g b)
  (declare (type card8 r g b))
  (rgb-set *vb-current-color* r g b)
  (vec4-set (material-diffuse *current-material*)
	    (/ r 255.0)
	    (/ g 255.0)
	    (/ b 255.0)
	    1.0))


(defconstant +begin-drawing-keys+ '(:POINTS :LINES :LINE-STRIP :LINE-LOOP :TRIANGLES :TRIANGLE-STRIP :TRIANGLE-FAN))
(defconstant +begin-drawing-funcs+ '(draw-points draw-lines draw-line-strip draw-line-loop draw-triangles draw-triangle-strip
						 draw-triangle-fan))


(defmacro Begin (mode &rest body)
  "Begin block holds rendering primitives"
  ;first, check if mode is valid
  (when (not (member mode +begin-drawing-keys+))
    (error "Begin: unknown primitive type: ~S" mode))

  (let ((bad-form nil))
    (labels ((body-form-valid-p (form)
				(case (car form)
				  ((Color Normal Vertex ArrayElement declare) t)
				  ((if when dotimes let let*) (body-valid-p (cddr form)))
				  ((do) (body-valid-p (cdddr form)))
				  (t (progn (setf bad-form form)
					    nil))))
	     (body-valid-p (body)
			   (if (null body) t
			     (and (body-form-valid-p (car body)) (body-valid-p (cdr body))))))
      ;second, check for unwanted function calls in the body
      (when (not (body-valid-p body))
	(error "Begin: illegal primitive call in form: ~S" bad-form))
      `(labels ((flush-vb ()
			  (do-transforms)
			  (when *do-normals-normalize* (do-normals-normalize))
			  (when *do-lighting* (do-lighting))
			  ;find the appropriate drawing function based on mode
			  ,(list (elt +begin-drawing-funcs+ (position mode +begin-drawing-keys+))))
		      
		(Normal (nx ny nz)
			(declare (type single-float nx ny nz))
			(vec3-set *vb-current-normal* nx ny nz)
			t)
		
		(Vertex (x y z w)
			(declare (type single-float x y z w))

					;if the vb is full, flush it
			(when (>= *vb-index* *vb-size*)
			  (progn (flush-vb)
				 (clear-vb)))
					;accumulate the vertex at the end of the vb    
			(set-vb-vertex *vb-index* :coords x y z w)
			(set-vb-vertex *vb-index* :normal
				       #{*vb-current-normal* 0} #{*vb-current-normal* 1} #{*vb-current-normal* 2})
			(set-vb-vertex *vb-index* :material *current-material*)
			(set-vb-vertex *vb-index* :color (rgb-r *vb-current-color*)
				       (rgb-g *vb-current-color*)
				       (rgb-b *vb-current-color*))
			(incf *vb-index*)))

	 (declare (inline Normal))
	 (clear-vb)
	 ,@body
	 (flush-vb)))))
     

;;; Vertex arrays stuff

(defparameter *vertex-array* nil)
;(declaim (type (simple-vector *) *vertex-array*))


(defmacro VertexPointer (array)
  `(setf *vertex-array* ,array))


(defmacro ArrayElement (i)
  (let ((index (gensym)))
    `(let ((,index (* 4 (the card32 ,i))))
       (declare (type card32 ,index))
       (Vertex (aref (the simple-array (eval *vertex-array*)) ,index)
	       (aref (the simple-array (eval *vertex-array*)) (incf ,index))
	       (aref (the simple-array (eval *vertex-array*)) (incf ,index))
	       (aref (the simple-array (eval *vertex-array*)) (incf ,index))))))


(defmacro DrawElements (mode count indices)
  `(Begin ,mode
	 (dotimes (i ,count)
	   (declare (type (simple-array card16 (*)) ,indices))
	   (ArrayElement (aref ,indices i)))))
	 

(defmacro DrawArrays (mode first count)
  `(Begin ,mode
	  (dotimes (i ,count)
	    (ArrayElement (+ ,first i)))))
