;;; ----------------------------------------------------------------------------
;;; $RCSfile: ASCParser.lisp,v $
;;; $Author: hubert $
;;; $Date: 2000/05/05 06:42:56 $
;;; $Revision: 1.2 $
;;; ----------------------------------------------------------------------------

;;GLOS (Graphic Library in Open Source), an ANSI Common Lisp OpenGL subset.
;;Copyright (C) 2000 the GLOS development team (http://glos.sourceforge.net)

(declaim (optimize (speed 0)
		   (compilation-speed 0)
		   (safety 0)
		   (debug 3)))


;;;-----------------------------------------------------------------------------
;;; Parser Constants
;;;-----------------------------------------------------------------------------


(defconstant +code-name+ "Named object:")
(defconstant +code-vx-num+ "Vertices:")
(defconstant +code-fa-num+ "Faces:")
(defconstant +code-vx-list+ "Vertex list:")
(defconstant +code-fa-list+ "Face list:")
(defconstant +code-smoothgroup+ "Smoothing:")
(defconstant +code-vx+ "Vertex ")
(defconstant +code-fa+ "Face ")
(defconstant +code-vx-coords+ '("X:" "Y:" "Z:"))
(defconstant +code-fa-kwds+ '("A:" "B:" "C:" "AB:" "BC:" "CA:"))


;;;-----------------------------------------------------------------------------
;;; Parser functions
;;;-----------------------------------------------------------------------------


(defun asc-parse-file (filename)
  "Run through the given ASC file and returns a list of objects"
  (let ((path (make-pathname :name filename)))
    (with-open-file
     (stream path :direction :input)
     (do* ((current-line "")		;current line of the stream
	   (object-list '()))
	 ((equal current-line nil) (nreverse object-list))
       (multiple-value-bind (object line) (asc-read-object stream current-line)
	 (setf current-line line)
	 (push object object-list))))))
       	 
	 
(defun asc-read-object (stream current-line)
  "Read the current open stream and its current line.
Returns the next object as 1st value, the new current line as second value.
Side-effect on the stream."
  (let* ((name nil)
	 (vertex-num nil)
	 (face-num nil)
	 (smoothgrp 0)
	 (vertex-array nil)		
	 (face-array nil)
	 (result nil)			;result of last multiline-seek
	 (single-seek (lambda (pat) (setf current-line
					  (stream-seek pat stream current-line))))
	 (multiline-seek
	  (lambda (x) (let* ((pat (mkstr +code-fa+ x ":"))
			     (stop-pat-lst (list (mkstr +code-fa+ (+ x 1) ":") +code-name+)))
			(multiple-value-bind (c-line results)
			    (stream-multiline-seek pat stop-pat-lst stream current-line)
			  (setf current-line c-line) results))))
	 (fetch-sf-cline (lambda (comname) (str-coerce-sf (str-fetch-arg comname current-line))))
	 (fetch-r-result (lambda (comname) (let ((x (str-fetch-arg comname result)))
					     (when x (str-coerce-r x))))))

    
    (block read-object
      ;;recuperate the minimal object ID      
      (ignore-errors
	(funcall single-seek +code-name+)
	(setf name (string-upcase (str-fetch-arg +code-name+ current-line)))
	(funcall single-seek +code-vx-num+)
	(setf vertex-num (str-coerce-r (str-fetch-arg +code-vx-num+ current-line)))
	(setf face-num (str-coerce-r (str-fetch-arg +code-fa-num+ current-line))))
	
      ;;exit now if the minimal information cannot be fetched       
      (when (not (and name vertex-num face-num))
	(return-from read-object (values nil current-line)))
      
      ;;get an array of vertices	      
      (when (> vertex-num 0)
	(setf vertex-array (vertex-array-make vertex-num))
	(funcall single-seek +code-vx-list+)
	(dotimes (i vertex-num)
	  (funcall single-seek (mkstr +code-vx+ i ":"))
	  (setf (aref vertex-array i)
		(multiple-value-bind (x y z)
		    (list-values (mapcar fetch-sf-cline +code-vx-coords+))
		  (vec4-make x y z 1.0)))))
	     
      ;;get an array of faces
      (when (> face-num 0)
	(setf face-array (face-array-make face-num))
	(funcall single-seek +code-fa-list+)
	(dotimes (i face-num)
	  (setf result (funcall multiline-seek i))
	  (setf smoothgrp (funcall fetch-r-result +code-smoothgroup+))
	  (unless smoothgrp (setf smoothgrp 0))
	  (setf (aref face-array i)
		(multiple-value-bind (a b c ab bc ca)
		    (list-values (mapcar fetch-r-result +code-fa-kwds+))
		  (fac-make :A a :B b :C c :AB ab :BC bc :CA ca :smoothing smoothgrp)))))

      ;;return the fetched object and the current stream-line       
      (values (obj-make :name name
			:vertices vertex-array
			:vertex-number vertex-num
			:faces face-array
			:face-number face-num)
	      current-line))))


;;;-----------------------------------------------------------------------------
;;; Functions generating "Begin" expressions from objects
;;;-----------------------------------------------------------------------------
   
(defun Begin-object (obj)
  (declare (optimize (speed 3)
		     (compilation-speed 0)
		     (safety 0)
		     (debug 0)))
  (Begin :TRIANGLES
	 (let ((nbface (obj-face-number obj)))
	   (declare (type fixnum nbface))
	   (do ((i 0 (1+ i)))
	       ((>= i nbface))
	     (declare (type fixnum i))
	     (let* ((face (aref (obj-faces obj) i))
		    (ia (fac-a face))
		    (ib (fac-b face))
		    (ic (fac-c face))
		    ;;(vf (aref (obj-fnormals obj) i))
		    (pa (aref (obj-vertices obj) ia))
		    (pb (aref (obj-vertices obj) ib))
		    (pc (aref (obj-vertices obj) ic))
		    (va (aref (obj-vnormals obj) ia))
		    (vb (aref (obj-vnormals obj) ib))
		    (vc (aref (obj-vnormals obj) ic)))
	       (declare (type vec4 pa pb pc)
			(type vec3 va vb vc))
	       ;;(Normal #{vf 0} #{vf 1} #{vf 2})
	       (Normal #{va 0} #{va 1} #{va 2})
	       (Vertex #{pa 0} #{pa 1} #{pa 2} #{pa 3})
	       (Normal #{vb 0} #{vb 1} #{vb 2})
	       (Vertex #{pb 0} #{pb 1} #{pb 2} #{pb 3})
	       (Normal #{vc 0} #{vc 1} #{vc 2})
	       (Vertex #{pc 0} #{pc 1} #{pc 2} #{pc 3}))))))

(defun load-list-asc (filename)
  (let ((obj-list (asc-parse-file filename)))
    (mapcar #'compute-normals obj-list)
    obj-list))
