;;; ----------------------------------------------------------------------------
;;; $RCSfile: Math3D.lisp,v $
;;; $Author: hubert $
;;; $Date: 2000/05/05 06:13:39 $
;;; $Revision: 1.1.1.1 $
;;; ----------------------------------------------------------------------------

;;GLOS (Graphic Library in Open Source), an ANSI Common Lisp OpenGL subset.
;;Copyright (C) 2000 the GLOS development team (http://glos.sourceforge.net)

(declaim (optimize (speed 3)
		   (compilation-speed 0)
		   (safety 0)
		   (debug 0)))


(defconstant +PI+ 3.141592653)
(declaim (type single-float +PI+))


(defconstant +mat44-index-list+ '((0 0) (0 1) (0 2) (0 3)
				  (1 0) (1 1) (1 2) (1 3)
				  (2 0) (2 1) (2 2) (2 3)
				  (3 0) (3 1) (3 2) (3 3)))


;;;-----------------------------------------------------------------------------
;;;-----------------------------------------------------------------------------
;; 4x4 Matrices 
;;;-----------------------------------------------------------------------------
;;;-----------------------------------------------------------------------------


(deftype mat44 () '(simple-array single-float (16)))


;; Constructor
(defun mat44-make-null ()
  "Create the null matrix"
  (make-array 16 :element-type 'single-float))
	       

;;;-----------------------------------------------------
;;; Matrix specific functions
;;;-----------------------------------------------------


; i and j must be numbers!
(defmacro mat44-elem (m i j)
  (declare (type card8 i j))
  `(the single-float (aref ,m ,(+ (* 4 i) j))))


(defmacro make-mat44-defuns ()
  "define accessors for the mat44 type"
  (let* ((index +mat44-index-list+)
	 (mat-access (lambda (mat ilist) `(#[,mat ,(car ilist) ,(cadr ilist)])))
	 (set-zero (lambda (x) `(setf ,@x 0.0)))
	 (accessors (mapcar (curry mat-access 'm) index))	
	 (mtarget (mapcar (curry mat-access 'mtarget) index))
	 (msource (mapcar (curry mat-access 'msource) index))
	 (mat-slots (mapcar (lambda (ilist) (symb 'M (car ilist) (cadr ilist)))
			    index)))
    `(progt (defun mat44-set-null (m)
	      "Modify a matrix which become a null matrix"
	      (declare (type mat44 m))
	      (progt ,@(mapcar set-zero accessors)))
	    
	    (defun mat44-copy (mtarget msource)
	      "copy the matrix msource into mtarget"
	      (declare (type mat44 m))
	      (progt ,@(mapcar (lambda (x y) `(setf ,@x ,@y))
			       mtarget msource)))

	    (defun mat44-set (m ,@mat-slots)
	      "fill a matrix with elements"
	      (progt ,@(mapcar (lambda (x y) `(setf ,@x ,y))
			       accessors mat-slots))))))


;;effective definition of the accessors in the project
(make-mat44-defuns)


(defun mat44-make-dup (m)
  "duplicate a matrix"
  (declare (type mat44 m))
  (copy-seq m))


(defun mat44-row (matrix rownum)
  "return a list, copy of the nth row of the given matrix"
  (generate-list 4
		 (lambda (x) #[matrix rownum x])
		 (j 0 (1+ j))))


(defun mat44-column (matrix colnum)
  "return a list, copy of the nth column of the given matrix"
  (generate-list 4
		 (lambda (x) #[matrix x colnum])
		 (i 0 (1+ i))))


(defun mat44-index-list ()
  "generate the list of indexes for the mat44 matrix type"
  (do ((i 0 (1+ i))
       (acc '()))
      ((>= i 4) (nreverse acc))
    (declare (type card8 i)) 
    (dotimes (j 4)
      (push `(,i ,j) acc))))


;;pretty-printing facility for displaying mat44 objects
(set-pprint-dispatch
 'mat44
 #'(lambda (s mat)
     (format s "~%~W~%~W~%~W~%~W
GLOS-MATRIX[4*4] OF SINGLE-FLOAT~%"
	     (mat44-row mat 0)
	     (mat44-row mat 1)
	     (mat44-row mat 2)
	     (mat44-row mat 3))))
	     

;;;-----------------------------------------------------------------------------
;;;-----------------------------------------------------------------------------
;; Vectors 
;;;-----------------------------------------------------------------------------
;;;-----------------------------------------------------------------------------


;;;-----------------------------------------------------------------------------
;; Constructors and Accessors
;;;-----------------------------------------------------------------------------

;We have both 3 and 4 elements vector types
;Functions or macros names prefixed with "vec-" take normally vec3 arguments. But for commodity,
;if a vec4 is passed the fourth coordinate w is ignored and the vec4 is treated as a vec3.

(deftype vec3 () '(simple-array single-float (3)))
(deftype vec4 () '(simple-array single-float (4)))


;; Constructors  
(defun vec3-make-null ()
  (make-array 3 :element-type 'single-float))


(defun vec3-make (x y z)
  (declare (type single-float x y z))
  (make-array 3 :element-type 'single-float :initial-contents (list x y z)))


(defun vec4-make-null ()
  (make-array 4 :element-type 'single-float))


(defun vec4-make (x y z w)
  (declare (type single-float x y z w))
  (make-array 4 :element-type 'single-float :initial-contents (list x y z w)))


;;;-----------------------------------------------------------------------------
;; Vector arithmetic operations
;;;-----------------------------------------------------------------------------

(declaim (inline vec3-set vec3-cpy
		 vec4-set vec4-cpy))

(defun vec3-set (v x y z)
  "Set coordinates of a vec3"
  (declare (type vec3 v)
	   (type single-float x y z))
  (setf #{v 0} x
	#{v 1} y
	#{v 2} z)
  t)


(defun vec3-cpy (v1 v2)
  "copy v2 (src) to v1 (dst)"
  (declare (type vec3 v1 v2))
  (setf #{v1 0} #{v2 0}
	#{v1 1} #{v2 1}
	#{v1 2} #{v2 2})
  t)


(defun vec4-set (v x y z w)
  "Set coordinates of a vec4"
  (declare (type vec4 v)
	   (type single-float x y z w))
  (setf #{v 0} x
	#{v 1} y
	#{v 2} z
	#{v 3} w)
  t)


(defun vec4-cpy (v1 v2)
  "copy v2 (src) to v1 (dst)"
  (declare (type vec4 v1 v2))
  (setf #{v1 0} #{v2 0}
	#{v1 1} #{v2 1}
	#{v1 2} #{v2 2}
	#{v1 3} #{v2 3})
  t)


;;these accessors are macros because we want them 'setable'
(defmacro vec-x (v)
  `(aref ,v 0))
(defmacro vec-y (v)
  `(aref ,v 1))
(defmacro vec-z (v)
  `(aref ,v 2))
(defmacro vec-w (v)
  `(aref ,v 3))


(defmacro vec-norm (v)
  "Return the length of the argument vector"
  `(the single-float (sqrt (the single-float (+ (* #{,v 0} #{,v 0})
						(* #{,v 1} #{,v 1})
						(* #{,v 2} #{,v 2}))))))


(defmacro vec-dot-product (v1 v2)
  "Return the dot product of the argument vectors"
  `(the single-float (+ (* #{,v1 0} #{,v2 0})
			(* #{,v1 1} #{,v2 1})
			(* #{,v1 2} #{,v2 2}))))


;;pretty-printing facility for displaying vec objects
(set-pprint-dispatch
 'vec3
 #'(lambda (s vec)
     (format s "~%{ ~W; ~W; ~W }
GLOS-VECTOR-3 OF SINGLE-FLOAT~%"
	     (vec-x vec)
	     (vec-y vec)
	     (vec-z vec))))


(set-pprint-dispatch
 'vec4
 #'(lambda (s vec)
     (format s "~%{ ~W; ~W; ~W; ~W }
GLOS-VECTOR-4 OF SINGLE-FLOAT~%"
	     (vec-x vec)
	     (vec-y vec)
	     (vec-z vec)
	     (vec-w vec))))


(defmacro vec-set-cross-product (vtarget v1 v2)
  "vtarget become the result of the cross product between v1 and v2"
  `(vec3-set ,vtarget
	     (- (* #{,v1 1} #{,v2 2}) (* #{,v1 2} #{,v2 1}))
	     (- (* #{,v1 2} #{,v2 0}) (* #{,v1 0} #{,v2 2}))
	     (- (* #{,v1 0} #{,v2 1}) (* #{,v1 1} #{,v2 0}))))


(defmacro vec-set-normalize (v)
  "normalized the vector v and return v"
  (let ((n (gensym)))
    `(let ((,n (vec-norm ,v)))
       (declare (type single-float ,n))
       (if (not (= ,n 0.0))
	   (vec3-set ,v (/ #{,v 0} ,n) (/ #{,v 1} ,n) (/ #{,v 2} ,n))))))
