;;; ----------------------------------------------------------------------------
;;; $RCSfile: Tools.lisp,v $
;;; $Author: hubert $
;;; $Date: 2000/05/05 06:13:39 $
;;; $Revision: 1.1.1.1 $
;;; ----------------------------------------------------------------------------

;;GLOS (Graphic Library in Open Source), an ANSI Common Lisp OpenGL subset.
;;Copyright (C) 2000 the GLOS development team (http://glos.sourceforge.net)

(declaim (optimize (speed 3)
		   (compilation-speed 0)
		   (safety 0)
		   (debug 0)))


;;;-----------------------------------------------------------------------------
;;; Useful Macros
;;;-----------------------------------------------------------------------------


;;Paul Graham's 'Ansi Common Lisp'
(defmacro while (test &body body)
  `(do ()
       ((not ,test))
     ,@body))


(defmacro progt (&body exps)
  "execute given expressions in sequence and returns T at the end."
  `(progn ,@exps
	  t))


;;arguments declaration & update observe the "do" syntax
;;Example: (generate-list 5 (lambda (x y) (+ x y))
;;                        (a 0 (+ a 2)) (b 1 (- b 1)))
;; --> (1 2 3 4 5)
(defmacro generate-list (len genfunc &body paramlist)
  "generate a list of "len" elements, result of the applying the given
lambda to a set of possibly updated arguments"
  (let ((params (mapcar #'car paramlist))
        (k (gensym))
        (acc (gensym)))
    `(do* (,@paramlist
           (,k 0 (1+ ,k))
           (,acc '()))
         ((>= ,k ,len) (reverse ,acc))
       (push (funcall #',genfunc ,@params) ,acc))))


;;Example: (setf myfun (lambda (x y z)
;;	      (format t "~%x:~A y:~A z:~A" x y z)))
;;(funcall (lcurry myfun 1 nil 3) 2)
;; --> x:1 y:2 z:3
(defmacro lcurry (lambdaform &body args)  
  "returns a curried version of given lambda"
  (let ((myargs '())
	(symlist '()))
    (progn (dolist (x args)
	     (if x
		 (push x myargs)
	       (let ((sym (gensym)))
		 (push sym symlist)
		 (push sym myargs))))
	   `(lambda ,(nreverse symlist)
	      (funcall ,lambdaform ,@(nreverse myargs))))))


;;consider multiple-value-setq for multiple-value assignment
(defmacro nsetf (value &body vars)
  "declare n variables initialized to an unique value."
  (let ((acc '()))
    (dolist (x vars)
      (push x acc)
      (push value acc))
    `(setf ,@(nreverse acc))))


(defmacro list-match-setf (lsrc ldest)
  "setf n variables to n values, all included in separate lists"
  (let ((gencode (lambda (x y) `(setf ,x ,y))))
    `(progn ,@(mapcar gencode lsrc ldest)
	    t)))


(defmacro list-match-incf (lsrc ldest)
  "incf n variables of n increments, all included in separate lists"
  (let ((gencode (lambda (x y) `(incf ,x ,y))))
    `(progn ,@(mapcar gencode lsrc ldest)
	    t)))


;;;-----------------------------------------------------------------------------
;;; Useful functions
;;;-----------------------------------------------------------------------------

(defun list-values (l)
  "return values of a list"
  (apply #'values l))

;;Paul Graham's 'Ansi Common Lisp'
(defun curry (fn &rest args)
  #'(lambda (&rest args2)
      (apply fn (append args args2))))

;;Paul Graham's 'On Lisp'
(defun mkstr (&rest args)
  (with-output-to-string (s)
			 (dolist (a args) (princ a s))))

;;Paul Graham's 'On Lisp'
(defun symb (&rest args)
  (values (intern (apply #'mkstr args))))


;;;-----------------------------------------------------------------------------
;;; Reader Macros
;;;-----------------------------------------------------------------------------


;; reader macro for accessing mat44 elements
;; #[m 2 3] -> 1.0
(progn
  (set-macro-character
   #\] (get-macro-character #\)))

  (set-dispatch-macro-character
   #\# #\[
   #'(lambda (stream char1 char2)
       (declare (ignore char1 char2))
       (let* ((trike (read-delimited-list #\] stream t))
	      (m (car trike))
	      (i (cadr trike))
	      (j (caddr trike)))
	 (if (numberp i) 
 	     (if (numberp j)
		 (let ((index (the fixnum (+ (coerce j 'fixnum) (* (coerce i 'fixnum) 4)))))
		   `(the single-float (aref ,m ,index)))
	       ; j n'est pas un nobre
	       (let ((index (the fixnum (* (coerce i 'fixnum) 4))))
		 `(the single-float (aref ,m (+ ,j ,index)))))
	   `(the single-float (aref ,m (+ ,j (the fixnum (* ,i 4))))))))))



;; reader macro for accessing vector slots
(progn
  (set-macro-character
   #\} (get-macro-character #\)))
  
  (set-dispatch-macro-character
   #\# #\{
   #'(lambda (stream char1 char2)
       (declare (ignore char1 char2))
       (let* ((trike (read-delimited-list #\} stream t))
	      (v (car trike))
	      (i (cadr trike)))
	 `(the single-float (aref ,v ,i))))))


;;;-----------------------------------------------------------------------------
;;; Useful utilities
;;;-----------------------------------------------------------------------------


;; multiple subst in a list:
;; substn '(1 a 2 b 3 c) '(a b 4 c coucou) -> '(1 2 4 3 coucou)
(defun substn (lst body)
  "multiple substitutions in a list"
  (if (null lst)
      body
    (substn (cddr lst)
	    (subst (first lst)
		   (second lst)
		   body))))



